
import unittest
import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization import Analyzer
from job_visualization import SynonymSet
from job_visualization.yaml_unfolder import YamlUnfolder

class TestParametersPresent(unittest.TestCase):
    def setUp(self):
        self.test_data = 'test/test_analysis/parameters_present/'

        self.yaml_unfolder = YamlUnfolder(root=self.test_data)

    def testParametersPresent(self):
        synonyms = SynonymSet()
        synonyms.add_set({'same-node', 'node-parameters'})

        arguments = ["parameters_present:same-node"]

        analyzer = Analyzer(root=self.test_data, synonyms=synonyms, arguments=arguments, file_index=self.yaml_unfolder.file_index, dep_extractor=self.yaml_unfolder.dep_extractor)
        analyzer.run()

        self.assertEqual(len(analyzer.results), 1)

        result = analyzer.results[0]

        self.assertEqual(result.errors[0].caller, 'publisher')
        self.assertEqual(result.errors[0].edge.to, 'test')
        self.assertEqual(result.errors[0].parameter, 'same-node')




import collections

from .result import Result

Error = collections.namedtuple('Error', ['caller', 'edge', 'parameter'])

def parameters_present(options, **kwargs):
    synonyms = options['synonyms']
    call_graph = options['call_graph']

    result = _Result()

    for node, edges in call_graph:
        for edge in edges:
            call_config = edge.settings

            for req_param, req_value in kwargs.items():
                found = False

                for param, value in call_config.items():
                    if synonyms.are_synonyms(param, req_param):
                        found = True
                        break

                if not found:
                    result.add(node, edge, req_param)

    return result

class _Result(Result):
    def add(self, node, edge, parameter):
        self.errors.append(Error(caller=node, edge=edge, parameter=parameter))

    def __str__(self):
        ret = "\nParameters present test\n---------\n"

        if len(self.errors) == 0:
            ret += "OK"
            return ret

        for error in self.errors:
            caller = error.caller
            edge = error.edge
            parameter = error.parameter
            ret += "{} calls {} without the required parameter {} (or synonyms)\n".format(caller, edge.settings['project'], parameter)

        return ret




from collections import OrderedDict
from os.path import splitext, exists
from os import makedirs

from .util import extract_from_config

def export_shell(file_index, to_dir):
    if not exists(to_dir):
        makedirs(to_dir)

    for _, file_data in file_index.files.items():
        export_from_file(file_data, to_dir)

def export_from_file(file_data, to_dir):
    shells = extract_shells_from_dict(file_data.yaml)
    path = file_data.path

    if len(shells) == 0:
        return

    for i in range(len(shells)):
        shell = shells[i]

        shell = extract_from_config(shell)

        if len(shells) == 1:
            file_name = "{}/{}.sh".format(to_dir, flatten_path(path))
        else:
            file_name = "{}/{}_{}.sh".format(to_dir, flatten_path(path), i)

        with open(file_name, 'w+') as f:
            f.write(shell)

def extract_shells_from_dict(yaml):
    shells = []

    if type(yaml) == OrderedDict or type(yaml) == dict:
        for key, value in yaml.items():
            if key == 'shell':
                shells.append(value)
            else:
                shells.extend(extract_shells_from_dict(value))

    elif type(yaml) == list:
        for value in yaml:
            shells.extend(extract_shells_from_dict(value))

    return shells

def flatten_path(path):
    path = path.replace('/', '-')

    path = splitext(path)[0]

    return path




import unittest
import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization import Analyzer
from job_visualization.yaml_unfolder import YamlUnfolder

class TestParametersPresent(unittest.TestCase):
    def setUp(self):
        self.test_data = 'test/test_analysis/cyclic_deps/'

        self.yaml_unfolder = YamlUnfolder(root=self.test_data)

    def testFindCycle(self):

        arguments = ['cyclic_deps']

        analyzer = Analyzer(root=self.test_data, arguments=arguments, file_index=self.yaml_unfolder.file_index, dep_extractor=self.yaml_unfolder.dep_extractor)
        analyzer.run()

        self.assertEqual(len(analyzer.results), 1)

        self.assertEqual(analyzer.results[0].errors[0], ['job3', 'job1', 'job2', 'job3'])





from collections import namedtuple, defaultdict

from .graphs import Edge

'''
project_name is a name of a job that is called
call_config is a config of call file, i.e. trigger-builds
project_config is a config of job that is been called
'''
class CallEdge(Edge):
    def __init__(self, to, call_config, project_config, caller_name):
        super(self.__class__, self).__init__(to, call_config)

        self.project_config = project_config
        self.caller_name = caller_name

'''
This flags are injected into unfolded configs so we can see that originaly they were included by
!include or !include-raw 
'''
include_flag = '__jenkins_job_builder_analysis_include_flag__'

'''
Info about included file
type is either 'include' or 'include-raw'
path is the original path to the file included
'''
IncludeInfo = namedtuple('IncludeInfo', ['type', 'path'])

 
class DepExtractor(object):

    def __init__(self, file_index, verbose=0):
        self.file_index = file_index
        self.verbose = verbose

    '''
    ==============
    Get calls part
    ==============
    '''

    def get_calls(self, job_name):
        '''
        Reads file by given name and returns CallEdge array
        '''

        config = self.file_index.get_by_name(job_name).yaml

        calls = self.get_calls_from_dict(config, from_name=job_name)

        return calls

    def get_calls_from_dict(self, file_dict, from_name, settings={}):
        '''
        Processes unfolded yaml object to CallEdge array

        settings is a dict of settings for keeping information like
        in what section we are right now (e.g. builders, publishers)
        '''

        calls = []
        call_settings = dict(settings)

        # Include all possible sections
        # The way to draw them is defined in call graph
        special_sections = {'builders', 'publishers', 'wrappers'}

        # Trigger flags
        triggers = {'trigger-builds', 'trigger-parameterized-builds'}

        if isinstance(file_dict, dict):
            for key in file_dict:

                if key in special_sections:
                    call_settings['section'] = key

                if key in triggers:
                    calls.extend(self.extract_call(file_dict[key], from_name, settings=call_settings))
                else:
                    calls.extend(self.get_calls_from_dict(file_dict[key], from_name, settings=call_settings))
        elif type(file_dict) == list:
            for value in file_dict:
                calls.extend(self.get_calls_from_dict(value, from_name, settings=call_settings))

        return calls


    def extract_call(self, call, from_name, settings):
        '''
        Creates CallEdge from call file (i.e. trigger-builds)

        Returns a list of calls
        '''

        call = defaultdict(lambda: None, call[0])

        try:
            call['section'] = settings['section']
        except KeyError:
            call['section'] = ''

        project = call['project']

        # If there is more than one call in a single file
        if type(project) == list:
            calls = []

            for name in project:
                calls.append(self.create_call(name, call, from_name))

            return calls
        else:
            return [self.create_call(project, call, from_name)]

    def create_call(self, project, call, from_name):
        file_data = self.file_index.get_by_name(project)

        call_object = CallEdge(to=project, call_config=call, project_config=file_data, caller_name=from_name)
        return call_object

    '''
    =================
    Get includes part
    =================
    '''

    def get_includes(self, path):
        config = self.file_index.unfold_yaml(path)

        return self.get_includes_from_dict(config, extract=True)

    def get_includes_from_dict(self, config, extract=False):
        if extract and isinstance(config, dict) and 'config' in config:
            config = config['config']
        
        includes = []

        if isinstance(config, dict):
            if include_flag in config:
                includes.append(config[include_flag])
            else:
                for key in config:
                    includes.extend(self.get_includes_from_dict(config[key]))
        elif isinstance(config, list):
            for value in config:
                includes.extend(self.get_includes_from_dict(value))

        return includes




from .file_index import FileIndex
from .yaml_unfolder import YamlUnfolder
from .file_data import FileData
from .config_parser import ConfigParser
from .analyzer import Analyzer
from .synonym_parser import SynonymSet
from .export_shell import export_shell
from .dep_extractor import DepExtractor

#import analysis
#import graphs




import graphviz as gv

class Legend(object):
    def __init__(self):
        # Legend is presented as map (name -> settings)
        self.items = {}

    def add_item(self, name, settings):
        self.items[name] = settings

    def render(self):
        legend = gv.Digraph('cluster_legend')
        legend.body.extend(['label="Legend"'])

        for name, settings in self.items.items():
            legend.node("{}-1".format(name), label="")
            legend.node("{}-2".format(name), label="")

            # format label so it doesn't overlap with edge
            label = "  {}".format(name)

            legend.edge("{}-1".format(name), "{}-2".format(name), label=label, **settings)

        legend_wrapper = gv.Digraph('cluster_legend_wrapper')
        legend_wrapper.subgraph(legend)

        legend_wrapper.body.extend(['style=invis'])

        return legend_wrapper



class SynonymSet:

    def __init__(self):
        self.synonyms = []

    def add_set(self, s):
        self.synonyms.append(s)

    def get_synonyms(self, param):
        for syn_set in self.synonyms:
            if param in syn_set:
                return syn_set

        return {}

    def are_synonyms(self, param_1, param_2):
        for syn_set in self.synonyms:
            if param_1 in syn_set and param_2 in syn_set:
                return True

        return False

    def __str__(self):
        return ':'.join(repr(syn_set) for syn_set in self.synonyms)

def parse_from_args(synonyms):
    '''
    Parse an array of string from argparser
    to SynonymSet
    '''

    syns_str = ''.join(synonyms)

    syns_str = syns_str.replace(' ', '')
    
    syn_set = SynonymSet()

    # to check if we are parsing inside the parenthesis
    inside_set = False
    current_syn = ''
    current_syn_set = set()

    for char in syns_str:
        if char == '{':
            inside_set = True
            current_syn_set = set()
            continue

        if char == '}':
            inside_set = False
            current_syn_set.add(current_syn)

            syn_set.add_set(current_syn_set)

            current_syn = ''

            continue

        if not inside_set:
            raise Exception("Incorrect synonyms {}".format(syns_str))

        if char == ',':
            if current_syn == '':
                raise Exception("Incorrect synonyms {}".format(syns_str))

            current_syn_set.add(current_syn)
            current_syn = ''

            continue

        current_syn += char

    return syn_set

def parse_from_array(arr):
    syn_set = SynonymSet()

    for synonyms in arr:
        _set = set()

        for synonym in synonyms:
            _set.add(synonym)

        syn_set.add_set(_set)

    return syn_set




from .result import Result

def cyclic_deps(options, **kwargs):
    call_graph = options['call_graph']

    call_result = _CallResult(cyclic_test(call_graph))
    
    return call_result


def cyclic_test(graph):
    cycles = []

    visited = set()

    for node, edges in graph:
        if node not in visited:

            cycle = find_cycle(graph, node, visited)

            if cycle is not None:
                cycles.append(cycle)

    cycles = remove_repetitions(cycles)

    return cycles

# Find one cycle that can be reached from the node
def find_cycle(graph, node, visited):
    current_stack = []

    return _find_cycle(graph, node, visited, current_stack)

# Simple DFS
# Current stack keeps path from the root to the current node
def _find_cycle(graph, node, visited, current_stack):
    visited.add(node)

    if node in current_stack:
        return unwrap_cycle(node, current_stack)

    current_stack.append(node)

    for edge in graph[node]:
        cycle = _find_cycle(graph, edge.to, visited, current_stack)

        if cycle is not None:
            return cycle

    current_stack.pop()

    return None

def unwrap_cycle(node, stack):
    cycle = [node]

    prev_node = stack.pop()

    while prev_node != node:
        cycle.append(prev_node)

        prev_node = stack.pop()

    cycle.append(node)

    cycle = cycle[::-1]

    return cycle

def remove_repetitions(cycles):
    seen = set()

    ret = []

    for cycle in cycles:
        cycle_set = frozenset(cycle)

        if cycle_set not in seen:
            ret.append(cycle)
            seen.add(cycle_set)

    return ret

def format_cycle(cycle):
    return " -> ".join(cycle)

class _CallResult(Result):
    def __init__(self, cycles):

        super(self.__class__, self).__init__()

        for cycle in cycles:
            self.add(cycle)

    def __str__(self):
        ret = "\nCyclic dependencies in call graph test\n-------------\n"

        if self.is_ok():
            ret += "OK"

            return ret

        for error in self.errors:
            ret += "Found cycle {}\n".format(format_cycle(error))

        return ret



import collections
from collections import OrderedDict


import graphviz as gv
import os

from . import graphs
from .file_index import FileIndex
from .file_data import FileData
from .dep_extractor import DepExtractor

  
class YamlUnfolder(object):

    def __init__(self, root, rank_dir=None, verbose=0):

        self.initing = True

        self.verbose = verbose
        self.root = root
        self.rank_dir = rank_dir

        self.init_file_index()

        self.dep_extractor = DepExtractor(self.file_index, verbose=verbose)

        self.init_include_graph()

        self.init_call_graph()

        self.initing = False

    def init_file_index(self):
        self.file_index = FileIndex(path=self.root, verbose=self.verbose)
        self.file_index.load_files(self.root)

    def init_include_graph(self):
        # Each graph should be able to have its own default rank_dir parameter
        if self.rank_dir is None:
            self.include_graph = graphs.include_graph.IncludeGraph(dep_extractor=self.dep_extractor, file_index=self.file_index)
        else:
            self.include_graph = graphs.include_graph.IncludeGraph(dep_extractor=self.dep_extractor, file_index=self.file_index, rank_dir=self.rank_dir)

    def init_call_graph(self):
        if self.rank_dir is None:
            self.call_graph = graphs.call_graph.CallGraph(dep_extractor=self.dep_extractor, file_index=self.file_index)
        else:
            self.call_graph = graphs.call_graph.CallGraph(rank_dir=self.rank_dir, dep_extractor=self.dep_extractor, file_index=self.file_index)

    def unfold_yaml(self, path):
        return self.file_index.unfold_yaml(path)

    def get_calls(self, path):
        return self.dep_extractor.get_calls(path)

    def get_includes(self, path):
        return self.dep_extractor.get_includes(path)





import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization.yaml_unfolder import YamlUnfolder
from test.graph_test import GraphTest


class TestLegend(GraphTest):

    def setUpYamlUnfolder(self, main_file):
        export_name = self.yaml_root + main_file + self.ext

        self.yaml_unfolder.call_graph.unfold_file(self.test_data + main_file)
        self.yaml_unfolder.call_graph.render(export_name)

    def setUp(self):
        self.test_data = 'test/test_graph/test_data/'
        self.test_refs = 'test/test_graph/test_refs/'

        self.yaml_root = self.test_data
        self.yaml_unfolder = YamlUnfolder(root=self.yaml_root)
        self.yaml_unfolder.call_graph.draw_legend = True

        self.ext = '_call'

    def testLegend(self):
        self.run_test_for_file('test_legend.yml')

if __name__ == "__main__":
    unittest.main()




import graphviz as gv

import collections
from ..file_data import FileData

from .graph import Graph
from .edge import Edge

from ..util import convert_path

class CallGraph(Graph):
    '''
    Class for manipulating call graph
    Stores all graph with file configs for further analysis
    '''
 
    def __init__(self, dep_extractor, file_index, rank_dir='left-right'):

        super(self.__class__, self).__init__(dep_extractor, file_index, rank_dir)

        self.active = False

        # Graphviz graph
        self.gv_graph = gv.Digraph(format='svg')

        if self.rank_dir == 'left-right':
            self.gv_graph.body.extend(['rankdir=LR'])

        self.gv_graph.body.extend(['size="8,5"'])

        # Internal graph represented as dict mapping node names to the list of its edges
        # Edge is represented as CallEdge
        self.graph = {}
        # Configs for files
        self._configs = {}
        # Roots are the jobs that were passed to `unfold_file` method
        # We need to treat them differently
        self._roots = set()

        # How to display call information
        # none - don't display
        # text - display as plain text above the edge
        # edge - display as node embeded in the edge
        self.call_display = 'text'

        # Which parameters to display on call edges
        self.call_parameters = {}

        self.init_legend()

    def init_legend(self):
        self.legend.add_item('publishers', {'color': 'green'})
        self.legend.add_item('builders', {'color': 'blue'})

    def add_call_object(self, call_object):
        self.add_node(call_object.to, call_object.project_config)
        self.add_edge(call_object.caller_name, call_object.to, call_object.settings)

    def add_node(self, name, project_config, is_root=False):
        if name not in self.graph:
            self.graph[name] = []
            self._configs[name] = project_config
            
        if is_root:
            self._roots.add(name)

    def add_edge(self, from_name, to_name, settings):
        if not self.has_edge(from_name, to_name):
            call_edge = Edge(to=to_name, settings=settings)
            self.graph[from_name].append(call_edge)

    def has_edge(self, from_name, to_name):

        try:
            edges = self.graph[from_name]
        except KeyError:
            return False

        for edge in edges:
            if edge.to == to_name:
                return True
        else:
            return False

    def unfold_file(self, path):
        yaml_config = self.file_index.unfold_yaml(path)

        self.unfold_config(path, yaml_config)

    def unfold_config(self, path, yaml_config):

        try:
            name = self.file_index.get_job_name(yaml_config)
        except KeyError:
            print("Warning: building call graph for not a job {}".format(yaml_config))
            return

        self.add_node(name, FileData(yaml=yaml_config, path=path), is_root=True)

        # Queue
        q = []

        q.extend(self.dep_extractor.get_calls(name))

        # What is the order of current call in call config
        current_order = 1

        # What is the name of the last caller
        # If the caller is the same, that means we are processing the call from the same caller
        # so we need to increase current_call
        current_caller = ''

        while len(q) != 0:
            call = q.pop(0)

            if call.caller_name != current_caller:
                current_order = 1
                current_caller = call.caller_name
            else:
                current_order += 1

            call.settings['call-order'] = current_order

            self.add_call_object(call)

            calls = self.dep_extractor.get_calls(call.to)

            for c in calls:
                if not self.has_edge(c.caller_name, c.to):
                    q.append(c)

    def render(self, path):
        for node in self._roots:
            self.render_node(node, color='red')

        for node in self.graph.keys():
            if node not in self._roots:
                self.render_node(node)

        # Because of the way GraphViz position clusters
        # we have to draw the base graph after the main graph
        super(self.__class__, self).render()

        self.gv_graph.render(path)


    def render_node(self, name, color='black'):
        self.gv_graph.node(self.get_path_from_name(name), color=color)

        edges = self.graph[name]

        for edge in edges:
            edge_settings = self.get_settings(edge)

            if self.call_display == 'none':
                self.render_simple_edge(name, edge, edge_settings)
            elif self.call_display == 'text':
                self.render_edge_with_label(name, edge, edge_settings)
            elif self.call_display == 'edge':
                self.render_edge_with_node_label(name, edge, edge_settings)
            else:
                raise Exception('Incorrect call display option {}'.format(self.call_display))

    def render_simple_edge(self, name, edge, edge_settings, label="call"):
        self.gv_graph.edge(self.get_path_from_name(name), self.get_path_from_name(edge.to), label=label, **edge_settings)

    def render_edge_with_label(self, name, edge, edge_settings):
        props_to_display = self.extract_props(edge.settings)

        label = '<'

        for prop, value in props_to_display.items():
            label += self.get_label(prop, value)
            label += "<BR/>"

        label += '>'

        self.gv_graph.edge(self.get_path_from_name(name), self.get_path_from_name(edge.to), label=label, **edge_settings)

    def render_edge_with_node_label(self, name, edge, edge_settings):
        props_to_display = self.extract_props(edge.settings)

        label = '<'
        label += "|".join(self.get_label(prop, value) for prop, value in props_to_display.items()) 
        label += '>'

        edge_node_name = "{}-{}".format(name, edge.to)

        self.gv_graph.node(edge_node_name, label=label, shape="record")

        self.gv_graph.edge(self.get_path_from_name(name), edge_node_name, arrowhead="none", **edge_settings)
        self.gv_graph.edge(edge_node_name, self.get_path_from_name(edge.to), **edge_settings)

    def get_label(self, prop, value):
        if value is None:
            return '{}: <FONT color="red">{}</FONT>'.format(prop, "not set")
        else:
            return "{}:{}".format(prop, value)

    def get_settings(self, edge):
        if 'section' not in edge.settings:
            return {}

        if edge.settings['section'] == 'publishers':
            return {'color': 'green'}

        if edge.settings['section'] == 'builders':
            return {'color': 'blue'}

        return {}

    def extract_props(self, settings):
        '''
        Extract all valuable properties to be displayed
        '''

        props = {}

        for param in self.call_parameters:
            if param in settings:
                props[param] = settings[param]
            else:
                props[param] = None

        return props

    def should_display_prop(self, prop):
        return prop in self.call_parameters

    def get_path_from_name(self, name):
        path = self._configs[name].path
        return convert_path(path)



from .parse import parse_analyzer_arguments
from .parameters_present import parameters_present
from .unused_configs import unused_configs
from .cyclic_deps import cyclic_deps
from .result import Result




import unittest
import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization import util

class TestConvertPath(unittest.TestCase):

    def testThrowsForAbsPath(self):
        with self.assertRaises(Exception) as context:
            util.convert_path('/abs/path')

        self.assertTrue('Cannot include' in str(context.exception))

    def testKeepsFileIfAlreadyCorrect(self):
        self.assertEquals('file.yml', util.convert_path('file.yml'))
        self.assertEquals('dir/file.yml', util.convert_path('dir/file.yml'))

    def testConvertsRelativePath(self):
        self.assertEquals('file.yml', util.convert_path('./file.yml'))
        self.assertEquals('dir/file.yml', util.convert_path('./dir/file.yml'))

    def testRemoveDuplicateSlashes(self):
        self.assertEquals('dir/file.yml', util.convert_path('dir//file.yml'))




import unittest
import abc
import os

class GraphTest(unittest.TestCase):

    def run_test_for_file(self, main_file):
        '''
        Unfolds main_file, renders it. Finds ref graph for it and compares generated graph with ref graph
        If ref graph is not found, generate it
        '''

        self.clean(main_file)

        self.setUpYamlUnfolder(main_file)

        with open(self.test_data + main_file + self.ext, 'r') as test_file:

            try:
                with open(self.test_refs + main_file + self.ext, 'r') as ref_file:
                    ref_graph = ref_file.read()
                    test_graph = test_file.read()

                    self.assertEquals(ref_graph, test_graph)
            except IOError:
                print('Generating new ref graph for ' + self.test_data + main_file)

                with open(self.test_refs + main_file + self.ext, 'w+') as ref_file:
                    test_graph = test_file.read()
                    ref_file.write(test_graph)

                    self.assertTrue(True)


    def clean(self, main_file):
        try:
            os.remove(self.test_data + main_file + self.ext)
            os.remove(self.test_data + main_file + self.ext + '.svg')
        except OSError:
            # Files were not created yet
            pass

    @abc.abstractmethod
    def setUpYamlUnfolder(self, main_file):
        pass




from .result import Result
from os.path import basename

def unused_configs(options, **kwargs):

    include_graph = options['include_graph']
    call_graph = options['call_graph']

    used_configs = get_used_configs(include_graph, call_graph)

    result = _Result()

    for config, edges in include_graph:
        if config not in used_configs and is_not_hidden(config): 
            result.add(config)

    return result

def get_used_configs(include_graph, call_graph):
    # Call graphs contains all the jobs. Assume all the jobs are used
    configs = set()

    for job, edges in call_graph:
        path_to_job = call_graph.get_path_from_name(job)
        configs.add(path_to_job)

    for node, edges in include_graph:
        for edge in edges:
            configs.add(edge.to)

    return configs

def is_not_hidden(config):
    return not basename(config).startswith('.')

class _Result(Result):

    def add(self, config):
        self.errors.append(config)

    def __str__(self):
        ret = "\nUnused configs test\n---------\n"

        if len(self.errors) == 0:
            ret += "OK"
            return ret

        for config in self.errors:
            ret += "{} is not used\n".format(config)

        return ret






import graphviz as gv

from .legend import Legend

class Graph(object):
    """
    Abstract class for render graphs
    """

    def __init__(self, dep_extractor, file_index, rank_dir):

        self.dep_extractor = dep_extractor
        self.file_index = file_index
        self.rank_dir = rank_dir

        self.legend = Legend()

        self.draw_legend = False

        self.graph = {}

    def render(self):
        if self.draw_legend:
            legend = self.legend.render()
            self.gv_graph.subgraph(legend)

    def __iter__(self):
        for node, edge in self.graph.items():
            yield node, edge

    def __getitem__(self, index):
        return self.graph[index]




import unittest
import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization.yaml_unfolder import YamlUnfolder
from job_visualization import FileIndex
from job_visualization import export_shell

class TestExportShell(unittest.TestCase):
    def setUp(self):
        self.test_data = 'test/test_export_shell/test_data'
        self.exported_shell = 'test/test_export_shell/exported_shell'

        self.yaml_unfolder = YamlUnfolder(root=self.test_data)

    def testExportShell(self):
        file_index = FileIndex(self.test_data)

        export_shell(file_index, to_dir=self.exported_shell)

        with open("{}/{}".format(self.exported_shell, "test-test_export_shell-test_data-job.sh"), "r") as f:
            self.assertEquals(f.read(), "echo test\necho second line\n")




import unittest
import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization import Analyzer
from job_visualization import SynonymSet
from job_visualization.yaml_unfolder import YamlUnfolder

class TestUnusedConfigs(unittest.TestCase):
    def setUp(self):
        self.test_data = 'test/test_analysis/unused_configs/'

        self.yaml_unfolder = YamlUnfolder(root=self.test_data)

    def testUnusedConfigs(self):

        arguments = ["unused_configs"]

        analyzer = Analyzer(root=self.test_data, arguments=arguments, file_index=self.yaml_unfolder.file_index, dep_extractor=self.yaml_unfolder.dep_extractor)
        analyzer.run()

        result = analyzer.results[0]

        self.assertEquals(len(result.errors), 2)

        self.assertEquals(result.errors[0], 'test/test_analysis/unused_configs/unused_1.yaml')
        self.assertEquals(result.errors[1], 'test/test_analysis/unused_configs/unused_2.yml')




import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization.yaml_unfolder import YamlUnfolder
from test.graph_test import GraphTest

class TestCallDisplay(GraphTest):

    def setUpYamlUnfolder(self, main_file):
        export_name = self.yaml_root + main_file + self.ext

        self.yaml_unfolder.call_graph.unfold_file(self.test_data + main_file)
        self.yaml_unfolder.call_graph.render(export_name)


    def setUp(self):
        self.test_data = 'test/test_call_graph/test_data/'
        self.test_refs = 'test/test_call_graph/test_refs/'

        self.yaml_root = self.test_data
        self.yaml_unfolder = YamlUnfolder(root=self.yaml_root)
        self.ext = '_call'

        self.yaml_unfolder.call_graph.call_parameters = {'same-node'}

    def testNoneDisplay(self):
        self.yaml_unfolder.call_graph.call_display = 'none'
        self.run_test_for_file('params_call_none.yml')

    def testTextDisplay(self):
        self.yaml_unfolder.call_graph.call_display = 'text'
        self.run_test_for_file('params_call_text.yml')

    def testEdgeDisplay(self):
        self.yaml_unfolder.call_graph.call_display = 'edge'
        self.run_test_for_file('params_call_edge.yml')

    def testNoParametersDisplay(self):
        self.yaml_unfolder.call_graph.call_parameters = {}
        self.run_test_for_file('no_params.yml')

    def testMultipleParamsDisplay(self):
        self.yaml_unfolder.call_graph.call_parameters = {'same-node', 'node-parameters'}
        self.run_test_for_file('multiple_params.yml')

    def testCallOrder(self):
        self.yaml_unfolder.call_graph.call_parameters = {'call-order'}
        self.run_test_for_file('order_call.yml')

if __name__ == "__main__":
    unittest.main()




import collections

FunctionArguments = collections.namedtuple('FunctionArguments', ['function', 'arguments'])

argument_splitter = ":"

def parse_analyzer_arguments(arguments):

    rets = []

    for argument in arguments:
        args = argument.split(argument_splitter)

        # The first one is the function name
        func_name = args[0]

        # The rest is the args
        func_args = {}

        for arg in args[1:]:
            key, value = parse_arg(arg)

            func_args[key] = value

        rets.append(FunctionArguments(function=func_name, arguments=func_args))

    return rets

def parse_arg(arg):
    arg = arg.split("=")

    if len(arg) == 1:
        return arg[0], True

    try:
        return arg[0], float(arg[1])
    except ValueError:
        if arg[1].lower() == 'true':
            return arg[0], True

        if arg[1].lower() == 'false':
            return arg[0], False

        return arg[0], arg[1]




import os

from . import yaml_unfolder
from . import synonym_parser

class ConfigParser(yaml_unfolder.YamlUnfolder):
    default_config = '.jjv_config.yml'

    def __init__(self, config, verbose=0):
        yaml_unfolder.YamlUnfolder.__init__(self, root='.', verbose=0)
        #super(self.__class__, self).__init__(root='.')

        self.verbose = verbose

        self.include_graph.active = False
        self.call_graph.active = False

        if os.path.exists(config):
            if self.verbose == 2:
                print("Using config {}".format(config))

            v = self.unfold_yaml(config)
            self.args = self.unfold_yaml(config)['config']
        else:
            if config != ConfigParser.default_config:
                raise IOError("No such file {}".format(config))

            self.args = {}

    def merge_args(self, args):
        for arg, value in self.args.items():
            if arg == 'synonyms':
                value = synonym_parser.parse_from_array(value)

            setattr(args, arg, value)

        return args




from distutils.core import setup
from setuptools import find_packages

setup(
    name='jabba',
    packages=find_packages(exclude=['docs', 'doc', 'test']),
    version='0.1.1',
    description='Jenkins And Job Builder Analysis',
    author='Missingdays',
    author_email='rebovykin@gmail.com',
    url='https://github.com/OSLL/jenkins_job_builder_visualization',
    download_url='https://github.com/OSLL/jenkins_job_builder_visualization/arhive/0.1.1.tar.gz',
    keywords=['jenkins', 'job buider', 'visualization'],
    scripts=['bin/jabba'],
    classifiers=[]
)






import unittest
import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization.analysis import parse_analyzer_arguments as parse

class TestParseAnalyzerArguments(unittest.TestCase):

    def testEmptyInput(self):
        self.assertEqual(len(parse([])), 0) 

    def testSimpleInput(self):
        args = parse(["func"])

        self.assertEqual(len(args), 1)
        self.assertEqual(args[0].function, "func")
        self.assertEqual(args[0].arguments, {})

    def testComplexInput(self):

        args = parse(["func1:arg1=True:arg2=False", "func2:arg1:arg2=10:arg3=hello"])

        self.assertEqual(len(args), 2)
        self.assertEqual(args[0].function, "func1")
        self.assertEqual(args[1].function, "func2")

        self.assertEqual(args[0].arguments['arg1'], True)
        self.assertEqual(args[0].arguments['arg2'], False)

        self.assertEqual(args[1].arguments['arg1'], True)
        self.assertEqual(args[1].arguments['arg2'], 10)
        self.assertEqual(args[1].arguments['arg3'], 'hello')
 




import os

def convert_path(path):
    if os.path.isabs(path):
        raise Exception("Cannot include file with absolute path {}. Please use relative path instead".format((path)))

    path = os.path.normpath(path)

    return path
 
def is_job_config(config):
    try:
        if config['config']['job']['name'] is not None:
            return True
    except KeyError:
        return False
    except TypeError:
        return False
    except IndexError:
        return False

    return False

def extract_from_config(config):
    if isinstance(config, dict) and 'config' in config:
        return config['config']

    return config




import unittest
import sys
import os

sys.path.append("../")
sys.path.append("../../")

from job_visualization.synonym_parser import *

class TestSynonymParser(unittest.TestCase):

    def testAreSynonyms(self):
        s = SynonymSet()

        s.add_set({'a', 'b'})

        self.assertTrue(s.are_synonyms('a', 'b'))
        self.assertFalse(s.are_synonyms('a', 'c'))
        self.assertFalse(s.are_synonyms('b', 'c'))
        

    def testParseFromArgs(self):

        syns = parse_from_args(['{node-parameters, ', 'same-node}'])

        self.assertTrue(syns.are_synonyms('node-parameters', 'same-node'))

        syns = parse_from_args(['{node-parameters, ', 'same-node}', '{a,', ' b,', 'c}'])

        self.assertTrue(syns.are_synonyms('node-parameters', 'same-node'))
        self.assertTrue(syns.are_synonyms('a', 'b'))
        self.assertTrue(syns.are_synonyms('b', 'c'))
        self.assertTrue(syns.are_synonyms('c', 'a'))

        self.assertFalse(syns.are_synonyms('same-node', 'a'))
        self.assertFalse(syns.are_synonyms('same-node', 'b'))
        self.assertFalse(syns.are_synonyms('same-node', 'c'))

    def testParseFromArray(self):
        syns = parse_from_array([['node-parameters', 'same-node']])

        self.assertTrue(syns.are_synonyms('node-parameters', 'same-node'))

        syns = parse_from_array([['node-parameters', 'same-node'], ['a', 'b', 'c']])

        self.assertTrue(syns.are_synonyms('node-parameters', 'same-node'))
        self.assertTrue(syns.are_synonyms('a', 'b'))
        self.assertTrue(syns.are_synonyms('b', 'c'))
        self.assertTrue(syns.are_synonyms('c', 'a'))

        self.assertFalse(syns.are_synonyms('same-node', 'a'))
        self.assertFalse(syns.are_synonyms('same-node', 'b'))
        self.assertFalse(syns.are_synonyms('same-node', 'c'))






import os

from collections import OrderedDict
from collections import namedtuple

import yaml
from yaml import load, Loader, dump

import collections
from .file_data import FileData
from .util import is_job_config, convert_path

from .analysis.cyclic_deps import format_cycle
from .dep_extractor import IncludeInfo, include_flag

def ordered_constructor(loader, node):
    loader.flatten_mapping(node)
    pairs = loader.construct_pairs(node)
    return OrderedDict(pairs)

class FileIndex:
    """
    Indexes files for quick search by name or path
    """
    def __init__(self, path, verbose=0):
        self.path = path
        self.verbose = verbose

        # name -> config for quick look up by job name
        self.jobs = {}

        # Dict: name -> (path, yaml)
        self.files = {}

        self.init_yaml_loader()

        # Jobs that were unfolded while unfolding the  file
        # Prevents goind into infinite recursion
        self.seen = set()

    def init_yaml_loader(self):
        Loader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, ordered_constructor)
        Loader.add_constructor('!include:', self.include_constructor)
        Loader.add_constructor('!include-raw:', self.include_raw_constructor)
        Loader.add_constructor('!include', self.include_constructor)


    def load_files(self, path):
        if self.verbose == 2:
            print("Indexing {}".format(path))

        for filename in os.listdir(path):

            file_path = path + "/" + filename

            if os.path.isdir(file_path):
                self.load_files(file_path)
            elif filename.endswith(".yaml") or filename.endswith(".yml"):
                self.unfold_yaml(file_path)

    def add_file(self, path, yaml):
        if is_job_config(yaml):
            name = self.get_job_name(yaml)

            file_data = FileData(path=path, yaml=yaml)
            self.files[path] = file_data
            self.jobs[name] = file_data
        else:
            self.files[path] = FileData(path=path, yaml=yaml)


    def get_job_name(self, file_config):
        # Job is represented as list
        name = file_config['config']['job']['name']

        return name

    def get_by_name(self, name):
        try:
            return self.jobs[name]
        except KeyError:
            raise KeyError("No job with name {} found in {}".format(name, self.path))

    def get_by_path(self, path):

        path = convert_path(path)

        try:
            return self.files[path]
        except KeyError:
            raise KeyError("No file {} found in {}".format(path, self.path))


    def __contains__(self, key):
        return key in self.files

    def unfold_yaml(self, path):
        '''
        Unfolds file by given name

        Also adds all included files as nodes to include graph. To disable this,
        set include_graph.active = False before calling
        '''

        self.unfolding_stack = []

        return self._unfold_yaml(path)

    def _unfold_yaml(self, path):

        path = convert_path(path)

        try:
            file_data = self.get_by_path(path)

            config = file_data.yaml

            return config
        except KeyError:
            # Maybe we haven't unfolded it yet
            pass

        # Found cycle
        if path in self.unfolding_stack:
            cyclic_call = self.unfolding_stack[-1]

            cycle = self.unfolding_stack[self.unfolding_stack.index(cyclic_call):]

            cycle = [path] + cycle + [path]

            raise Exception("Cyclic include found. {}".format(format_cycle(cycle)))

        self.unfolding_stack.append(path)

        with open(path, 'r') as f:
            config = load(f)

            config = self.inject_include_info(path, config, include_type='include')

            self.add_file(path, config)

            self.unfolding_stack.pop()

            return config

    def inject_include_info(self, path, config, include_type):
        if isinstance(config, list):
            config = config[0]

        ret = OrderedDict()

        ret[include_flag] = IncludeInfo(type=include_type, path=path)
        ret['config'] = config

        return ret

    def include_constructor(self, loader, node):
        v = self._unfold_yaml(node.value)

        return v

    def include_raw_constructor(self, loader, node):

        path = convert_path(node.value)

        with open(path, 'r') as f:
            config = f.read()

            config = self.inject_include_info(path, config, include_type='include-raw')

            self.add_file(path, config)

            return config

    def default_load(self, name, config):
        pass




from .yaml_unfolder import YamlUnfolder
from . import graphs
from . import analysis
from .analysis import parse_analyzer_arguments
from .util import is_job_config
from .synonym_parser import SynonymSet

class Analyzer(YamlUnfolder):

    def __init__(self, root, arguments, file_index, dep_extractor, synonyms=SynonymSet(), verbose=0):
        super(self.__class__, self).__init__(root=root, verbose=0)

        self.verbose = verbose

        self.arguments = parse_analyzer_arguments(arguments)
        self.synonyms = synonyms
        self.root = root
        self.file_index = file_index

        self.include_graph = graphs.include_graph.IncludeGraph(dep_extractor=dep_extractor, file_index=file_index)
        self.include_graph.active = True
        self.call_graph = graphs.call_graph.CallGraph(dep_extractor=dep_extractor, file_index=self.file_index)

        self.results = []

        self.create_include_graph()
        self.create_call_graph()

    def create_include_graph(self):
        if self.verbose == 2:
            print("Creating full include graph for analysis")

        for name, file_data in self.file_index.files.items():
            self.include_graph.unfold_file(file_data.path)

    def create_call_graph(self):
        if self.verbose == 2:
            print("Creating full call graph for analysis")

        for name, file_data in self.file_index.files.items():
            if not is_job_config(file_data.yaml):
                continue

            self.call_graph.unfold_config(file_data.path, file_data.yaml)
        
    def run(self):
        self.results = []

        options = {
            'synonyms': self.synonyms,
            'include_graph': self.include_graph,
            'call_graph': self.call_graph,
            'file_index': self.file_index,
            'dep_extractor': self.dep_extractor
        }

        for argument in self.arguments:
            func = None

            try:
                func = getattr(analysis, argument.function)
            except AttributeError:
                raise AttributeError("Cannot find analysis function {}".format(argument.function))

            if self.verbose == 2:
                print("Running {} analysis function".format(argument.function))

            result = func(options=options, **argument.arguments)
            self.results.append(result)

    def is_ok(self):
        return all(result.is_ok() for result in self.results)

    def print_result(self):
        for result in self.results:
            print(result)




import graphviz as gv

import collections

from .graph import Graph
from .edge import Edge

'''
All visual settings we want to specify for include graph
'''
graph_settings = {
    'edges': {
        'include_color': 'green',
        'include_raw_color': 'red',
        'default_color': 'black'
    }
}

class IncludeGraph(Graph):

    def __init__(self, dep_extractor, file_index, rank_dir='left-right'):
        super(self.__class__, self).__init__(dep_extractor, file_index, rank_dir)

        self.gv_graph = gv.Digraph(format='svg')

        if self.rank_dir == 'left-right':
            self.gv_graph.body.extend(['rankdir=LR'])

        self.gv_graph.body.extend(['size="8,5"'])

        self.active = False

        # Internal graph
        # Represented as hashmap (path -> edges) where edges is list of IncludeEdge objects
        self.graph = {}

        self.init_legend()

    def init_legend(self):
        self.legend.add_item('include', {'color': 'green'})
        self.legend.add_item('include-raw', {'color': 'red'})


    def add_node(self, name):
        if not self.active:
            return

        if name not in self.graph:
            self.graph[name] = []

    def add_edge(self, node_from, node_to, type):
        if not self.active:
            return


        self.add_node(node_from)
        self.add_node(node_to)

        if self.has_edge(node_from, node_to):
            return

        edges = self.graph[node_from]

        if type == 'include':
            color = graph_settings['edges']['include_color']
            label='<<B>include</B>>'
        elif type == 'include-raw':
            color = graph_settings['edges']['include_raw_color']
            label='<<B>include-raw</B>>'
        else:
            color = graph_settings['edges']['default_color']
            label=''

        edges.append(Edge(to=node_to, settings = {
            'label': label,
            'color': color
        }))

    def has_edge(self, from_node, to_node):
        if not self.active:
            return False

        if not from_node in self.graph:
            return False

        edges = self.graph[from_node]

        for edge in edges:
            if edge.to == to_node:
                return True

        return False

    def unfold_file(self, path):
        yaml_config = self.file_index.unfold_yaml(path)

        self.unfold_config(path, yaml_config)

    def unfold_config(self, path, yaml_config):

        # Queue stores config files and who includes them
        q = [(path, None)]

        # To speed things up
        seen = set()

        while len(q) != 0:
            config_name, included_from = q.pop(0)

            if config_name in seen:
                continue

            seen.add(config_name)

            self.add_node(config_name)
            
            if included_from is not None:
                self.add_edge(included_from['name'], config_name, type=included_from['type'])

            includes = self.dep_extractor.get_includes(config_name)

            for include in includes:
                el = (include.path, {'name': config_name, 'type': include.type})
                q.append(el)

    def render(self, file_to):
        if not self.active:
            return

        for path in self.graph.keys():
            self.gv_graph.node(path)

            for edge in self.graph[path]:
                self.gv_graph.edge(path, edge.to, label=edge.settings['label'], color=edge.settings['color'])

        super(self.__class__, self).render()

        self.gv_graph.render(file_to)





import unittest
import sys
import os

from argparse import Namespace

sys.path.append("../")
sys.path.append("../../")

from job_visualization.config_parser import ConfigParser

class TestConfigParser(unittest.TestCase):

    def setUp(self):
        self.test_data = 'test/test_config_parser/test_data/'

    def testLoadFile(self):
        parser = ConfigParser(self.test_data + 'config_1.yml')

        self.assertEquals(parser.args['yaml_root'], 'root-directory')
        self.assertEquals(parser.args['legend'], True)
        self.assertEquals(parser.args['call_parameters'], ['same-node', 'node-parameters'])

    def testMergeArgs(self):
        parser = ConfigParser(self.test_data + 'config_1.yml')

        args = Namespace(
            yaml_root = 'old_dir',
            call_parameters=[],
            call_graph=True
        )

        args = parser.merge_args(args)

        self.assertEquals(args.yaml_root, 'root-directory')
        self.assertEquals(args.legend, True)
        self.assertEquals(args.call_parameters, ['same-node', 'node-parameters'])
        self.assertEquals(args.call_graph, True)

